#include "basicmaterial.h"

using namespace std;

namespace Msp {
namespace GL {

BasicMaterial::BasicMaterial():
	receive_shadows(false)
{
	set_diffuse(Color(1.0f));
	set_specular(Color(0.0f));
	set_emission(Color(0.0f));
	set_shininess(50.0f);
	set_reflectivity(0.0f);
}

string BasicMaterial::create_program_source() const
{
	string source = "import phong;\n";
	if(diffuse.texture)
		source += "const bool use_diffuse_map = true;\n";
	if(specular.texture || specular.value.r || specular.value.g || specular.value.b)
	{
		source += "const bool use_specular = true;\n";
		if(specular.texture)
			source += "const bool use_specular_map = true;\n";
		if(shininess.texture)
			source += "const bool use_shininess_map = true;\n";
	}
	if(normal.texture)
		source += "const bool use_normal_map = true;\n";
	if(emission.texture || emission.value.r || emission.value.g || emission.value.b)
	{
		source += "const bool use_emission = true;\n";
		if(emission.texture)
			source += "const bool use_emission_map = true;\n";
	}
	if(reflectivity.value || reflectivity.texture)
	{
		source += "const bool use_reflectivity = true;\n";
		if (reflectivity.texture)
			source += "const bool use_reflectivity_map = true;\n";
	}
	if(receive_shadows)
		source += "const bool use_shadow_map = true;\n";
	return source;
}

void BasicMaterial::attach_textures_to(Texturing &texturing, ProgramData &tex_shdata) const
{
	attach_texture_to(diffuse.texture, texturing, tex_shdata, "diffuse_map");
	attach_texture_to(specular.texture, texturing, tex_shdata, "specular_map");
	attach_texture_to(normal.texture, texturing, tex_shdata, "normal_map");
	attach_texture_to(emission.texture, texturing, tex_shdata, "emission_map");
	attach_texture_to(shininess.texture, texturing, tex_shdata, "shininess_map");
	attach_texture_to(reflectivity.texture, texturing, tex_shdata, "reflectivity_map");
}

void BasicMaterial::set_diffuse(const Color &color)
{
	diffuse.value = color;
	shdata.uniform("basic_material.diffuse", color);
}

void BasicMaterial::set_diffuse_map(const Texture *tex)
{
	diffuse.texture = tex;
}

void BasicMaterial::set_specular(const Color &color)
{
	specular.value = color;
	shdata.uniform("basic_material.specular", color);
}

void BasicMaterial::set_specular_map(const Texture *tex)
{
	specular.texture = tex;
}

void BasicMaterial::set_normal_map(const Texture *tex)
{
	normal.texture = tex;
}

void BasicMaterial::set_emission(const Color &color)
{
	emission.value = color;
	shdata.uniform("basic_material.emission", color);
}

void BasicMaterial::set_emission_map(const Texture *tex)
{
	emission.texture = tex;
}

void BasicMaterial::set_shininess(float value)
{
	shininess.value = value;
	shdata.uniform("basic_material.shininess", value);
}

void BasicMaterial::set_shininess_map(const Texture *tex)
{
	shininess.texture = tex;
}

void BasicMaterial::set_reflectivity(float value)
{
	reflectivity.value = value;
	shdata.uniform("basic_material.reflectivity", value);
}

void BasicMaterial::set_reflectivity_map(const Texture *tex)
{
	reflectivity.texture = tex;
}

void BasicMaterial::set_receive_shadows(bool s)
{
	receive_shadows = s;
}


DataFile::Loader::ActionMap BasicMaterial::Loader::shared_actions;

BasicMaterial::Loader::Loader(BasicMaterial &m):
	DerivedObjectLoader<BasicMaterial, Material::PropertyLoader<BasicMaterial> >(m)
{
	set_actions(shared_actions);
}

BasicMaterial::Loader::Loader(BasicMaterial &m, Collection &c):
	DerivedObjectLoader<BasicMaterial, Material::PropertyLoader<BasicMaterial> >(m, c)
{
	set_actions(shared_actions);
}

void BasicMaterial::Loader::init_actions()
{
	Material::PropertyLoader<BasicMaterial>::init_actions();
	add_property("diffuse", &BasicMaterial::set_diffuse, &BasicMaterial::set_diffuse_map, true);
	add_property("specular", &BasicMaterial::set_specular, &BasicMaterial::set_specular_map, false);
	add_property("normal", &BasicMaterial::set_normal_map);
	add_property("emission", &BasicMaterial::set_emission, &BasicMaterial::set_emission_map, false);
	add_property("shininess", &BasicMaterial::set_shininess, &BasicMaterial::set_shininess_map);
	add_property("reflectivity", &BasicMaterial::set_reflectivity, &BasicMaterial::set_reflectivity_map);
	add("receive_shadows", &BasicMaterial::receive_shadows);
}

} // namespace GL
} // namespace Msp
